"""\
Session Objects
"""
########################################################################
#
# Session.py,v 0.0.3b 2000/10/10
#
# Copyright (c) 2000-2001 Technology Concepts & Design, Inc.
# All rights reserved. Written by Corey Zimmet - caz@tcdi.com
#
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################



import string, sys, time, whrandom

import OFS.Folder
from OFS.Document import Document
from Globals import HTMLFile, Persistent, PersistentMapping
import Acquisition, AccessControl, OFS, string
import os
import re
import urllib
import whrandom
import tempfile
from DateTime import DateTime
import threading,time
from Shared.DC.ZRDB.TM import TM

import sys

_sesspatt=re.compile("^(/.*?)(/\$.*)$")
_nonsesspatt=re.compile("^(/.*?)(/.*)+$")
_generator=whrandom.whrandom()
_strAlphaNum="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890"
# Define load/dump functions



addSessionForm=HTMLFile('dtml/Session/addSession', globals())

def _randomID():
	#import pdb;pdb.set_trace()
	i=0
	id="S__"
	while i<8:
		id=id+_strAlphaNum[_generator.randint(0,61)]
		i=i+1
	return id


def addSession(self, id, title="", REQUEST=None):
    """Add a Session to Zope"""
    
    self._setObject(id, SessionManager(id, title))
    
    if REQUEST is not None: 
	# Display the folder that contains the new Session
	return self.manage_main(self,REQUEST)



class SessionInstance(Acquisition.Implicit):


	'''Session Object'''
	meta_type='SessionInstance'
	manage_options =OFS.SimpleItem.Item.manage_options

	__ac_permissions__=(
		('View management screens',('manage_tabs','manage_main','manage_displaySession')),
		('Change permissions',('manage_access',)),
		)
	manage_display=HTMLFile('dtml/SessionInstance/manage_displaySession',globals())
	__allow_access_to_unprotected_subobjects__=1
	#_permissionIsSet=None




	def __init__(self,id,timeout=30.0,title="A session instance"):
		self.id=id
		self.title=title
		
		self._v_data={}
		self._v_data['id']=id
		self._v_data['title']=title
		self.Session=self

		self.accessedTime=DateTime()
		self._creationTime=DateTime()
		self.setTimeOut(timeout)

	def _CurrentSession(self):
		"""return current session"""		
		return self

	def __getattr__(self, key):
		#import pdb;pdb.set_trace()
		try:
			r=self._v_data[key]
			return r
		except:
			raise AttributeError, key
	def __len__(self):
		return len(self._v_data)
	
	def __setitem__(self, key, item):
		self._v_data[key] = item
		
	def __delitem__(self, key):
		del self._v_data[key]
		
	def clear(self):
		self._v_data.clear()
		
	def keys(self):
		return self._v_data.keys()
	
	def items(self):
		return self._v_data.items()
	
	def values(self):
		return self._v_data.values()
	
	def has_key(self, key):
		return self._v_data.has_key(key)

	def __getitem__(self, key):
		return self._v_data.get(key)

	def setTimeOut(self,timeout):
		self.timeout=timeout
		self.expirationInterval=((1.0/24.0)/60.0)*float(timeout)

	def _expired(self):
		now=DateTime()
		expirationTime=self.accessedTime+self.expirationInterval
		if expirationTime<now:
			return 1
		else:
			return 0
	def creationTime(self):
		"""return creation time"""
		return DateTime(self._creationTime.timeTime())

	def _updateLastAccessedTime(self):
		self.accessedTime=DateTime()




	
	set=__setitem__
	get=__getitem__
	delete=__delitem__



class SessionManager(
	OFS.Folder.Folder
	):
	"""A Session
	"""

	icon='misc_/CookieLess/icon'
	_properties = (
		{'id':'title','type':'string','mode':'w'},
		{'id':'timeout','type':'float','mode':'w'}
		)

	meta_type='CookieLessSessionManager'	# Objects Name in the Zope list

	manage_options = OFS.Folder.Folder.manage_options+(
		{'label':'Sessions','icon':icon,'action':'manage_session','target':'manage_main'},
		)


	__ac_permissions__=(
		('View management screens',('manage_tabs',

									'manage_main',

									'manage_session',

									'sessions',

									'manage_displaySession',

									'getSession',

									'manage_clone',

									'manage_cutObjects',

									'manage_copyObjects',

									'manage_pasteObjects',

									'manage_renameForm',

									'manage_renameObject')),
		('Change permissions',('manage_access',)),
		)

	manage_session=HTMLFile('dtml/Session/manage_session',globals())
	manage_displaySession=HTMLFile('dtml/Session/manage_displaySession',globals())
	_sessions={}
	timeout=30.0



	def __init__(self, id, title='',startup=None):
		self.id = id
		self.title = title
		
		if startup==None:

			startup=_onSessionStartup

		self._setObject('onSessionStartup', Document(startup, __name__='onSessionStartup'))

		

	def sessions(self):
		return self._sessions.values()


	def _checkExpired(self):
		while 1:

			try:
				for session in self._sessions.keys():
					if self._sessions[session]!=None:
						if self._sessions[session]._expired():
							self._expireSession(session)
			except:

				pass
			time.sleep(60) # wake up every minute and check sessions

	def _expireSession(self,session):
		if self._sessions.has_key(session):
			del self._sessions[session]

	def manage_expireSessions(self,sessionlist,REQUEST=None):
		"""Expires selected sessions"""
		for session in sessionlist:
			self._expireSession(session)
		if REQUEST:
			return self.manage_session(self,REQUEST)

	def manage_expireAllSessions(self,REQUEST=None):
		"""Expires selected sessions"""
		self._sessions.clear()
		if REQUEST:
			return self.manage_session(self,REQUEST)

	def __bobo_traverse__(self, REQUEST, name=None):
		#import pdb;pdb.set_trace()

		name=urllib.unquote(name)
		bypassSession=None
		stack=REQUEST['TraversalRequestNameStack']

		if name[0:3]!="S__":
			if  name[:6]=="manage":
				bypassSession=1
			else:
				for itemfromstack in stack:
					if itemfromstack[:6]=='manage':
							bypassSession=1

			
		if bypassSession:
			if hasattr(self, 'aq_base'):
				b = self.aq_base
				if hasattr(b, name):
					return getattr(self, name)
				try:
					return self[name]
				except:
					return getattr(self, name)
				
		if len(name)< 3 or name[0:3] !="S__":
			if hasattr(REQUEST,'URL'):
				return self(REQUEST)
			else:
				return getattr(self,name)
		elif self._sessions.has_key(name):
			session = self._sessions[name]
			session._updateLastAccessedTime()
			session=session.__of__(self)
			#if not session._permissionIsSet:
			#	session.manage_permission('View',['Anonymous'],1)
			return session
		else:
			raise 'SessionExpired', name

	def getSession(self,session):
		""" return a session's information """
		#import pdb;pdb.set_trace()
		if self._sessions.has_key(session):
			return self._sessions[session]
		else:
			raise 'SessionExpired', name

	def __call__(self,REQUEST):
		#import pdb;pdb.set_trace()
		import os.path
		import urlparse

		
		protocol,server,urlpath,params,qry,frags= urlparse.urlparse(REQUEST.URL)
		urlpath=REQUEST['PATH_INFO']
		if not _sesspatt.match(urllib.unquote(urlpath)):
			session=SessionInstance(_randomID(),self.timeout)

			if hasattr(self,'onSessionStartup'):

				REQUEST.set('Session',session)

				self.onSessionStartup(self,REQUEST)
			self._sessions[session.id]=session
			if _nonsesspatt.match(urlpath):
				urlpath= re.sub(_nonsesspatt,"\\1/%s\\2"%(session.id),urlpath)
			else:
				urlpath="%s/%s"%(urlpath,session.id)
			newurl=urlparse.urlunparse((protocol,server,urlpath,params,qry,frags))
			raise 'Redirect',newurl

	def __setstate__(self,state):
		'Set initial state'
		Persistent.__setstate__(self,state)
		if not hasattr(self,"_sessions"):
			self._sessions={}
		if not hasattr(self,"_v_thread"):
			self._v_thread=threading.Thread(None,self._checkExpired)
			self._v_thread.setDaemon(1)
			self._v_thread.start()





_onSessionStartup="""

<dtml-comment>

          Insert any code that you want to be executed on Session startup below:

</dtml-comment>

"""









